import pathlib
from typing import Any, Dict, Generator, Tuple

import jinja2

from main import app

VUE_TEMPLATE = """
<script setup lang="ts">
  #{ if in == 'query' }
const @{ name } = ref < @{ type } > (@{ default });
  #{ else }
const @{ name } = ref < @{ type } > (@{ default });
  #{ endif }

  #{ if body }
  export interface @{ body.content['application/json'].schema.title } {
    #{ for k, v in body.content['application/json'].schema.properties.items() }
  @{ k }: @{ v.type };
    #{ endfor }
}
  #{ endif }

const url = computed(() => {
  let url = '@{ path }' + '?' + @{ params }.map(p => p.name + '=' + p.value).join('&');
  return url;
});

const trigger = ref(false);
</script>

<template>
  <Request method="@{ method }" :url="url" :headers="headers" :body="body" v-if="trigger">
    <slot :json="{ json }"></slot>

    <slot #actions="{ json }">
      <button @click="trigger = !trigger">Send</button>
    </slot>

    <slot #error="{ json }">
      <div v-if="json.error">
        <h3>Error</h3>
        <pre>{{ json.error }}</pre>
      </div>
    </slot>
  </Request>
</template>
"""


class Templater(object):
    def generator(self) -> Generator[Dict[str, Any], None, None]:
        """
        Generator that yields a dictionary of data for each endpoint.
        on the OpenAPI spec.
        """
        paths = app.openapi["paths"]
        for path in paths:
            for k, v in paths[path].items():
                method = k.upper()
                name = v["summary"]
                _body = v.get("requestBody")
                _params = v["parameters"]
                params = []
                for param in _params:
                    _in = param["in"]
                    name = param["name"]
                    required = param["required"]
                    _type = param["schema"]["type"]
                    default = param["schema"].get("default")

                    yield {
                        "path": path,
                        "method": method,
                        "in": _in,
                        "required": required,
                        "default": default,
                        "type": _type.__name__,
                        "name": name,
                        "params": params,
                        "body": _body,
                        "headers": {"Content-Type": "application/json"},
                    }

    @property
    def template(self) -> str:
        """The Base Vue Template with an slotted Request component."""
        template_str = (
            pathlib.Path(__file__)
            .parent.joinpath("templates", "__init__.vue")
            .read_text()
        )
        return template_str

    def render(self) -> Generator[Tuple[Any,str], None, None]:
        """
        Generator that yields a rendered template for each endpoint.
        on the OpenAPI spec.
        """
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchpath="templates"),
            variable_start_string="@{",
            variable_end_string="}",
            comment_start_string="##{",
            comment_end_string="}",
            block_start_string="#{",
            block_end_string="}",
        )
        template = env.from_string(self.template)
        for data in self.generator():
            yield data.get("name"), template.render(**data)


def generate_components() -> None:
    """Generate Vue Components for each endpoint on the OpenAPI spec."""
    for name, component in Templater().render():
        if not pathlib.Path("app/src/components").exists():
            pathlib.Path("app/src/components").mkdir(parents=True)
        pathlib.Path(f"app/src/components/{name}.vue").write_text(component)  
        print(f"Generated {name}.vue")
        
        
def main():
  number_of_components = len(list(Templater().generator()))
  print(f"{number_of_components} will be generated are you sure you want to continue? (y/n)")
  if input() == "y":
    generate_components()
    print("Done!")
  else:
    print("Aborted")
  
  
if __name__ == "__main__":
  main()


import asyncio
import functools
import os
import pathlib
import signal
import socket
import subprocess
import sys
import traceback
from typing import Any

import click
from aiohttp.web import run_app
from aiohttp_devtools import __version__
from aiohttp_devtools.cli import cli
from aiohttp_devtools.exceptions import AiohttpDevException
from aiohttp_devtools.logs import main_logger, setup_logging
from aiohttp_devtools.runserver import INFER_HOST
from aiohttp_devtools.runserver import runserver as _runserver
from aiohttp_devtools.runserver import serve_static

from .build import main as build_main
from .curl import main as curl_main
from .git import main as git_main
from .kill import main as kill_main
from .vue import main as vue_main

_dir_existing = click.Path(exists=True, dir_okay=True, file_okay=False)
_file_dir_existing = click.Path(exists=True, dir_okay=True, file_okay=True)
_dir_may_exist = click.Path(
    dir_okay=True, file_okay=False, writable=True, resolve_path=True
)


verbose_help = "Enable verbose output."
livereload_help = (
    "Whether to inject livereload.js into html page footers to autoreload on changes. "
    "env variable AIO_LIVERELOAD"
)


static_help = "Path of static files to serve, if excluded static files aren't served. env variable: AIO_STATIC_STATIC"
root_help = (
    "Root directory project used to qualify other paths. "
    "This can be used to watch a parent directory for changes in a more complex application. "
    "env variable: AIO_ROOT"
)
static_url_help = 'URL path to serve static files from, default "/static/". env variable: AIO_STATIC_URL'
host_help = (
    "host used when referencing livereload and static files, if blank host is taken from the request header "
    "with default of localhost. env variable AIO_HOST"
)
app_factory_help = (
    "name of the app factory to create an aiohttp.web.Application with, if missing default app-factory "
    "names are tried. This can be either a function with signature "
    '"def create_app(loop): -> Application" or "def create_app(): -> Application" '
    "or just an instance of aiohttp.Application. env variable AIO_APP_FACTORY"
)
port_help = "Port to serve app from, default 8000. env variable: AIO_PORT"
aux_port_help = "Port to serve auxiliary app (reload and static) on, default port + 1. env variable: AIO_AUX_PORT"

def gen_port() -> int:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port


@click.group()
@click.version_option(__version__, "-V", "--version")
@click.option("-v", "--verbose", is_flag=True, help=verbose_help)
@click.pass_context
def main(ctx: click.Context, verbose: bool) -> None:
    """
    AioFauna CLI
    """
    setup_logging(verbose=verbose)
    ctx.obj = {}

    if verbose:
        main_logger.debug("Verbose mode enabled")


# defaults are all None here so default settings are defined in one place: DEV_DICT validation
@main.command()
@click.argument(
    "app-path",
    envvar="AIO_APP_PATH",
    type=_file_dir_existing,
    required=False,
    default=".",
)
@click.option(
    "-s",
    "--static",
    "static_path",
    envvar="AIO_STATIC_PATH",
    type=_dir_existing,
    help=static_help,
    default="./static",
)
@click.option(
    "--root",
    "root_path",
    envvar="AIO_ROOT",
    type=_dir_existing,
    help=root_help,
    default=".",
)
@click.option(
    "--static-url", envvar="AIO_STATIC_URL", help=static_url_help, default="/"
)
@click.option(
    "--livereload/--no-livereload",
    envvar="AIO_LIVERELOAD",
    default=True,
    help=livereload_help,
)
@click.option("--host", default=INFER_HOST, help=host_help)
@click.option(
    "-p",
    "--port",
    "main_port",
    envvar="AIO_PORT",
    type=click.INT,
    help=port_help,
    default=8080,
)
@click.option(
    "--app-factory", "app_factory_name", envvar="AIO_APP_FACTORY", help=app_factory_help
)
@click.option(
    "--aux-port",
    envvar="AIO_AUX_PORT",
    type=click.INT,
    help=aux_port_help,
    default=gen_port(),
)
@click.option("-v", "--verbose", is_flag=True, help=verbose_help, default=True)
@click.argument("project_args", nargs=-1)
def dev(**config: Any) -> None:
    """
    Run a development server for aiofauna/aiohttp apps.

    Takes one argument "app-path" which should be a path to either a directory containing a recognized default file
    ("app.py" or "main.py") or to a specific file. Defaults to the environment variable "AIO_APP_PATH" or ".".

    The app path is run directly, see the "--app-factory" option for details on how an app is loaded from a python
    module.
    """
    active_config = {k: v for k, v in config.items() if v is not None}
    setup_logging(config["verbose"])
    # Rewrite argv for the application.
    sys.argv[1:] = active_config.pop("project_args")
    try:
        run_app(**_runserver(**active_config))
    except AiohttpDevException as e:
        if config["verbose"]:
            tb = click.style(traceback.format_exc().strip("\n"), fg="white", dim=True)
            main_logger.warning("AiohttpDevException traceback:\n%s", tb)
        main_logger.error("Error: %s", e)
        sys.exit(2)


@main.command()
def prod():
    """
    Run a production server for aiofauna/aiohttp apps.
    """
    subprocess.check_output(["python", "main.py"])
    

@main.command()
def build():
    """
    Build project
    """
    build_main()


@main.command()
def git():
    """
    Push to git
    """
    git_main()


@main.command()
def kill():
    """
    Kill port
    """
    kill_main()
    
@main.command()
def curl():
    """
    Run curl
    """
    curl_main()
    
@main.command()
def components():
    """
    Generate components
    """
    vue_main()


import os
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod


class BuildStrategy(ABC):
    @abstractmethod
    def build(self) -> None:
        pass


UNDEBUGGABLE_EXIT_CODES = (1, 2, 126, 127, 128, 130, 255)


class NodeJSBinaryStrategy(BuildStrategy):
    def build(self) -> None:
        try:
            platform = sys.platform
            node = self._search_node_binary(platform)
            npm = self._search_npm_binary(platform)
            print(f"Node: {node}")
            print(f"NPM: {npm}")
            spa_project_path: str = "./app"
            subprocess.check_call([node, "--version"])
            subprocess.check_call([npm, "--version"])
            subprocess.check_call([npm, "install"], cwd=spa_project_path)
            subprocess.check_call([npm, "run", "build"], cwd=spa_project_path)
        except Exception as e:
            print(e)
            raise e

    def _search_node_binary(self, platform: str) -> str:
        if platform == "linux":
            _ref = subprocess.check_output(["which", "node"])
            return _ref.decode("utf-8").strip()
        if platform == "darwin":
            _ref = subprocess.check_output(["which", "node"])
            return _ref.decode("utf-8").strip()
        if platform == "win32":
            _ref = subprocess.check_output(["where", "node"])
            return _ref.decode("utf-8").strip()
        raise Exception("Unsupported platform: {}".format(platform))

    def _search_npm_binary(self, platform: str) -> str:
        if platform == "linux":
            _ref = subprocess.check_output(["which", "npm"])
            return _ref.decode("utf-8").strip()
        if platform == "darwin":
            _ref = subprocess.check_output(["which", "npm"])
            return _ref.decode("utf-8").strip()
        if platform == "win32":
            _ref = subprocess.check_output(["where", "npm"])
            return _ref.decode("utf-8").strip()
        else:
            try:
                self._forcefully_install_nodejs()

            except OSError as e:
                _exit_code_number = e.errno

                if _exit_code_number in UNDEBUGGABLE_EXIT_CODES:
                    self._gracefully_exit()
                else:
                    raise e
            except Exception as e:
                raise e
            else:
                self._search_npm_binary(platform)
        raise Exception("Unsupported platform: {}".format(platform))

    def _forcefully_install_nodejs(self) -> None:
        try:
            subprocess.check_call(["sudo", "apt", "install", "nodejs", "-y"])
        except Exception as e:
            raise e

    def _gracefully_exit(self) -> None:
        print("NodeJS is not installed. Please install it and try again.")
        sys.exit(1)


class PythonStrategy(BuildStrategy):
    def build(self) -> None:
        try:
            subprocess.check_call(
                ["python", "-m", "pip", "install", "-r", "requirements.txt"]
            )
            if not os.path.exists("templates"):
                os.makedirs("templates")
                shutil.copy("static/index.html", "templates/index.html")
        except Exception as e:
            raise e


class Build:
    def __init__(self, strategy: BuildStrategy) -> None:
        self.strategy = strategy

    def build(self) -> None:
        self.strategy.build()

    def __call__(self) -> None:
        self.strategy.build()


def build_node() -> None:
    build = Build(NodeJSBinaryStrategy())
    build()     
    
    
def build_python() -> None:
    build = Build(PythonStrategy())
    build()
    
    
def main() -> None:
    build_node()
    build_python()
    


import hashlib
import os
import subprocess


def run_command(command):
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
    process.wait()


def git_add():
    run_command("git add .")


def git_commit(path: str = "."):
    sha = hashlib.sha256()
    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith(".py"):
                with open(os.path.join(root, file), "rb") as f:
                    sha.update(f.read())
                    print(sha.hexdigest())
        for path in dirs:
            git_commit(path)

    run_command(f'git commit -m "{sha.hexdigest()}"')


def git_push():
    run_command("git push origin main")


def main():
    git_add()
    git_commit()
    git_push()



"""
Flaskaesque helper functions for aiohttp.
"""
import json
import re
from functools import singledispatch
from typing import Any, Dict, List, Optional, Union

from aiohttp import web
from aiohttp.web import Response
from jinja2 import Environment, FileSystemLoader, select_autoescape
from pygments import highlight
from pygments.formatters import \
    HtmlFormatter  # pylint: disable=no-name-in-module
from pygments.lexers import get_lexer_by_name

from markdown import markdown  # pylint: disable=import-error

from .exceptions import HttpException
from .odm import BaseModel, FaunaModel

MD_FMT_CDN = """<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/github-markdown-css/github-markdown.min.css">"""  # pylint: disable=line-too-long
CODE_FMT_CDN = """<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/pygments-css@1.0.0/github.min.css">"""  # pylint: disable=line-too-long

html_formatter = HtmlFormatter(style="default")

jinja_env = Environment(
    loader=FileSystemLoader("templates"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
    extensions=["jinja2.ext.do"],
)

md_env = Environment(
    loader=FileSystemLoader("markdown"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
)


def markdown_it(md_string: str):
    """
    Render a markdown file to HTML.

    Args:
        md_string (str): The markdown file to render.

    Returns:
        str: The rendered HTML.
    """

    template = md_env.get_template(md_string)
    text = template.render()
    code = re.findall(r"```[a-z]*\n[\s\S]*?\n```", text)
    for code_ in code:
        lang = re.findall(r"```([a-z]*)\n", code_)[0]
        lexer = get_lexer_by_name(lang)
        code = re.findall(r"```[a-z]*\n([\s\S]*?)\n```", code_)[0]
        code = highlight(code, lexer, html_formatter)
        text = text.replace(code_, code)
    text = markdown(text, extensions=["fenced_code"])
    text = f"""
    <html>
        <head>
            <meta charset="utf-8">
            {MD_FMT_CDN}
            {CODE_FMT_CDN}
        </head>
        <style>
        .container {{
            max-width: 800px;
            margin: 0 auto;
            padding: 20px;
        }}
        h1, h2, h3, h4, h5, h6 {{
            margin-top: 40px;
            text-align: center;
            font-weight: 600;
            font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Helvetica, Arial, sans-serif;
        }}
        </style>
        <body class="markdown-body">
        <main class="container">
            {text}
        </main>
        </body>
    </html>
    """
    return Response(text=text, content_type="text/html")


def redirect(url):
    """
    Create an HTTP redirect response.

    Args:
        url (str): The URL to redirect to.

    Returns:
        web.HTTPFound: An HTTP redirect response with the specified URL.
    """

    return web.HTTPFound(url)


def make_response(data, headers=None, status=200):
    """
    Create an HTTP response with the given data, headers, and status code.

    Args:
        data (str): The response body data.
        headers (dict, optional): A dictionary of headers to include in the response. Defaults to None.
        status (int, optional): The HTTP status code for the response. Defaults to 200.

    Returns:
        web.Response: An HTTP response with the specified data, headers, and status code.
    """
    return web.Response(body=data, headers=headers, status=status)


def render_template(template_name, **kwargs):
    """
    Render an HTML template with the given context variables and return an HTTP response.

    Args
        template_name (str): The name of the template file.
        **kwargs: Keyword arguments representing context variables to be passed to the template.

    Returns:
        web.Response: An HTTP response with the rendered template as its body and content type set to "text/html".
    """
    template = jinja_env.get_template(template_name).render(**kwargs)
    return web.Response(body=template.encode(), content_type="text/html")


@singledispatch
def do_response(response: Any) -> Response:
    """
    Flask-esque function to make a response from a function.
    """
    return response


@do_response.register(BaseModel)
def _(response: BaseModel) -> Response:
    return Response(
        status=200, body=response.json(
            exclude_none=True, exclude_unset=True
            ), content_type="application/json"
    )


@do_response.register(FaunaModel)
def _(response: FaunaModel) -> Response:
    response.ref = str(response.ref)
    return Response(
        status=200, body=response.json(
            exclude_none=True, exclude_unset=True
            ), content_type="application/json"
    )


@do_response.register(dict)
def _(response: dict) -> Response:
    return Response(
        status=200, body=json.dumps(response), content_type="application/json"
    )


@do_response.register(str)
def _(response: str) -> Response:
    return Response(status=200, body=response.encode(), content_type="text/plain")


@do_response.register(bytes)
def _(response: bytes) -> Response:
    return Response(status=200, body=response, content_type="application/octet-stream")


@do_response.register(int)
def _(response: int) -> Response:
    return Response(status=200, body=str(response).encode(), content_type="text/plain")


@do_response.register(float)
def _(response: float) -> Response:
    return Response(status=200, body=str(response).encode(), content_type="text/plain")


@do_response.register(bool)
def _(response: bool) -> Response:
    return Response(status=200, body=str(response).encode(), content_type="text/plain")


@do_response.register(list)
def _(response: List[Union[FaunaModel, BaseModel, dict, str, int, float]]) -> Response:
    processed_response = []

    for item in response:
        if issubclass(item.__class__, BaseModel):
            processed_response.append(item.dict())  # type: ignore
        else:
            processed_response.append(item)

    return Response(
        status=200, body=json.dumps(processed_response), content_type="application/json"
    )

@do_response.register(HttpException)
def _(response: HttpException) -> Response:
    return Response(
        status=response.status_code, body=json.dumps(response.json()), content_type="application/json"
    )



import os
import signal
import subprocess


def kill_process_on_port(port):
    command = f"lsof -t -i:{port}"
    pid_str = subprocess.check_output(command, shell=True)
    if isinstance(pid_str, bytes):
        pid_str = pid_str.decode("utf-8")
    elif isinstance(pid_str, str):
        pass
    else:
        raise TypeError(f"Expected bytes or string, got {type(pid_str)}")

    if pid_str:
        for pid in pid_str.split("\n"):
            if pid:  # Check if pid is not an empty string
                os.kill(int(pid), signal.SIGKILL)
                print(f"Killed process {pid} on port {port}")
    else:
        print(f"No process on port {port}")


def main():
    port = input("Enter port number to kill process: ")
    kill_process_on_port(port)



from typing import Mapping, Optional
from .._types import TypedDict


__all__ = (
    'PartialModel',
    'PartialModelField',
)


class PartialModelField(TypedDict):
    name: str
    is_list: bool
    optional: bool
    type: str
    documentation: Optional[str]
    is_relational: bool


class PartialModel(TypedDict):
    name: str
    from_model: str
    fields: Mapping[str, PartialModelField]




from __future__ import annotations

import json
import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional

import click

from .. import config
from ..errors import PrismaError
from ._node import node, npm

log: logging.Logger = logging.getLogger(__name__)


def run(
    args: List[str],
    check: bool = False,
    env: Optional[Dict[str, str]] = None,
) -> int:
    log.debug('Running prisma command with args: %s', args)

    default_env = {
        **os.environ,
        'PRISMA_HIDE_UPDATE_MESSAGE': 'true',
        'PRISMA_CLI_QUERY_ENGINE_TYPE': 'binary',
    }
    env = {**default_env, **env} if env is not None else default_env

    # TODO: ensure graceful termination
    entrypoint = ensure_cached().entrypoint
    process = node.run(
        str(entrypoint),
        *args,
        env=env,
        check=check,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )

    if args and args[0] in {'--help', '-h'}:
        click.echo(click.style('Python Commands\n', bold=True))
        click.echo(
            '  '
            + 'For Prisma Client Python commands run '
            + click.style('prisma py --help', bold=True)
        )

    return process.returncode


class CLICache(NamedTuple):
    cache_dir: Path
    entrypoint: Path


DEFAULT_PACKAGE_JSON: dict[str, Any] = {
    'name': 'prisma-binaries',
    'version': '1.0.0',
    'private': True,
    'description': 'Cache directory created by Prisma Client Python to store Prisma Engines',
    'main': 'node_modules/prisma/build/index.js',
    'author': 'RobertCraigie',
    'license': 'Apache-2.0',
}


def ensure_cached() -> CLICache:
    cache_dir = config.binary_cache_dir
    entrypoint = cache_dir / 'node_modules' / 'prisma' / 'build' / 'index.js'

    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)

    # We need to create a dummy `package.json` file so that `npm` doesn't try
    # and search for it elsewhere.
    #
    # If it finds a different `package.json` file then the `prisma` package
    # will be installed there instead of our cache directory.
    package = cache_dir / 'package.json'
    if not package.exists():
        package.write_text(json.dumps(DEFAULT_PACKAGE_JSON))

    if not entrypoint.exists():
        click.echo('Installing Prisma CLI')

        try:
            proc = npm.run(
                'install',
                f'prisma@{config.prisma_version}',
                cwd=config.binary_cache_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if proc.returncode != 0:
                click.echo(
                    f'An error ocurred while installing the Prisma CLI; npm install log: {proc.stdout.decode("utf-8")}'
                )
                proc.check_returncode()
        except Exception:
            # as we use the entrypoint existing to check whether or not we should run `npm install`
            # we need to make sure it doesn't exist if running `npm install` fails as it will otherwise
            # lead to a broken state, https://github.com/RobertCraigie/prisma-client-py/issues/705
            if entrypoint.exists():
                try:
                    entrypoint.unlink()
                except Exception:
                    pass
            raise

    if not entrypoint.exists():
        raise PrismaError(
            f'CLI installation appeared to complete but the expected entrypoint ({entrypoint}) could not be found.'
        )

    return CLICache(cache_dir=cache_dir, entrypoint=entrypoint)



import builtins
import copy
import logging
import operator
import re
from configparser import ConfigParser
from typing import Any, Callable, Dict, Optional
from typing import Type as TypingType
from typing import Union, cast

from mypy.errorcodes import ErrorCode
from mypy.nodes import (BytesExpr, CallExpr, Context, DictExpr, Expression,
                        IntExpr, NameExpr, Node, StrExpr, SymbolTable,
                        SymbolTableNode, TypeInfo, Var)
from mypy.options import Options
from mypy.plugin import CheckerPluginInterface, MethodContext, Plugin
from mypy.types import Instance, NoneType, Type, UnionType

# match any direct children of an actions class
CLIENT_ACTION_CHILD = re.compile(
    r'prisma\.actions\.(.*)Actions\.(?P<name>(((?!\.).)*$))'
)
ACTIONS = [
    'create',
    'find_unique',
    'delete',
    'update',
    'find_first',
    'find_many',
    'upsert',
    'update_many',
    'delete_many',
    'count',
]

CONFIGFILE_KEY = 'prisma-mypy'

log: logging.Logger = logging.getLogger(__name__)


# due to the way the mypy API is typed we unfortunately have to disable Pyright type checks
# this is because mypy type hints are written like this: `Bogus[str]` instead of `str`
# mypy uses internal magic to transform Bogus[T] to T which pyright cannot understand.
# pyright: reportGeneralTypeIssues=false, reportUnnecessaryComparison=false


def plugin(version: str) -> TypingType[Plugin]:
    return PrismaPlugin


class PrismaPluginConfig:
    __slots__ = ('warn_parsing_errors',)
    warn_parsing_errors: bool

    def __init__(self, options: Options) -> None:
        if options.config_file is None:  # pragma: no cover
            return

        plugin_config = ConfigParser()
        plugin_config.read(options.config_file)
        for key in self.__slots__:
            setting = plugin_config.getboolean(
                CONFIGFILE_KEY, key, fallback=True
            )
            setattr(self, key, setting)


class PrismaPlugin(Plugin):
    config: PrismaPluginConfig

    def __init__(self, options: Options) -> None:
        self.config = PrismaPluginConfig(options)
        super().__init__(options)

    def get_method_hook(
        self, fullname: str
    ) -> Optional[Callable[[MethodContext], Type]]:
        match = CLIENT_ACTION_CHILD.match(fullname)
        if not match:
            return None

        if match.group('name') in ACTIONS:
            return self.handle_action_invocation

        return None

    def handle_action_invocation(self, ctx: MethodContext) -> Type:
        # TODO: if an error occurs, log it so that we don't cause mypy to
        #       exit prematurely.
        return self._handle_include(ctx)

    def _handle_include(self, ctx: MethodContext) -> Type:
        """Recursively remove Optional from a relational field of a model
        if it was explicitly included.

        An argument could be made that this is over-engineered
        and while I do agree to an extent, the benefit of this
        method over just setting the default value to an empty list
        is that access to a relational field without explicitly
        including it will raise an error when type checking, e.g

        user = await client.user.find_unique(where={'id': user_id})
        print('\n'.join(p.title for p in user.posts))
        """
        include_expr = self.get_arg_named('include', ctx)
        if include_expr is None:
            return ctx.default_return_type

        if not isinstance(ctx.default_return_type, Instance):
            # TODO: resolve this?
            return ctx.default_return_type

        is_coroutine = self.is_coroutine_type(ctx.default_return_type)
        if is_coroutine:
            actual_ret = ctx.default_return_type.args[2]
        else:
            actual_ret = ctx.default_return_type

        is_optional = self.is_optional_type(actual_ret)
        if is_optional:
            actual_ret = cast(UnionType, actual_ret)
            model_type = actual_ret.items[0]
        else:
            model_type = actual_ret

        if not isinstance(model_type, Instance):
            return ctx.default_return_type

        try:
            include = self.parse_expression_to_dict(include_expr)
            new_model = self.modify_model_from_include(model_type, include)
        except Exception as exc:
            log.debug(
                'Ignoring %s exception while parsing include: %s',
                type(exc).__name__,
                exc,
            )

            # TODO: test this, pytest-mypy-plugins does not bode well with multiple line output
            if self.config.warn_parsing_errors:
                # TODO: add more details
                # e.g. "include" to "find_unique" of "UserActions"
                if isinstance(exc, UnparsedExpression):
                    err_ctx = exc.context
                else:
                    err_ctx = include_expr

                error_unable_to_parse(
                    ctx.api,
                    err_ctx,
                    'the "include" argument',
                )

            return ctx.default_return_type

        if is_optional:
            actual_ret = cast(UnionType, actual_ret)
            modified_ret = self.copy_modified_optional_type(
                actual_ret, new_model
            )
        else:
            modified_ret = new_model  # type: ignore

        if is_coroutine:
            arg1, arg2, _ = ctx.default_return_type.args
            return ctx.default_return_type.copy_modified(
                args=[arg1, arg2, modified_ret]
            )

        return modified_ret

    def modify_model_from_include(
        self, model: Instance, data: Dict[Any, Any]
    ) -> Instance:
        names = model.type.names.copy()
        for key, node in model.type.names.items():
            names[key] = self.maybe_modify_included_field(key, node, data)

        return self.copy_modified_instance(model, names)

    def maybe_modify_included_field(
        self,
        key: Union[str, Expression, Node],
        node: SymbolTableNode,
        data: Dict[Any, Any],
    ) -> SymbolTableNode:
        value = data.get(key)
        if value is False or value is None:
            return node

        if isinstance(value, (Expression, Node)):
            raise UnparsedExpression(value)

        # we do not want to remove the Optional from a field that is not a list
        # as the Optional indicates that the field is optional on a database level
        if (
            not isinstance(node.node, Var)
            or node.node.type is None
            or not isinstance(node.node.type, UnionType)
            or not self.is_optional_union_type(node.node.type)
            or not self.is_list_type(node.node.type.items[0])
        ):
            log.debug(
                'Not modifying included field: %s',
                key,
            )
            return node

        # this whole mess with copying is so that the modified field is not leaked
        new = node.copy()
        new.node = copy.copy(new.node)
        assert isinstance(new.node, Var)
        new.node.type = node.node.type.items[0]

        if (
            isinstance(value, dict)
            and 'include' in value
            and isinstance(new.node.type, Instance)
            and isinstance(new.node.type.args[0], Instance)
        ):
            model = self.modify_model_from_include(
                new.node.type.args[0], value['include']
            )
            new.node.type.args = (model, *new.node.type.args)

        return new

    def get_arg_named(
        self, name: str, ctx: MethodContext
    ) -> Optional[Expression]:
        """Return the expression for an argument."""
        # keyword arguments
        for i, names in enumerate(ctx.arg_names):
            for j, arg_name in enumerate(names):
                if arg_name == name:
                    return ctx.args[i][j]

        # positional arguments
        for i, arg_name in enumerate(ctx.callee_arg_names):
            if arg_name == name and ctx.args[i]:
                return ctx.args[i][0]

        return None

    def is_optional_type(self, typ: Type) -> bool:
        return isinstance(typ, UnionType) and self.is_optional_union_type(typ)

    def is_optional_union_type(self, typ: UnionType) -> bool:
        return len(typ.items) == 2 and isinstance(typ.items[1], NoneType)

    # TODO: why is fullname Any?

    def is_coroutine_type(self, typ: Instance) -> bool:
        return bool(typ.type.fullname == 'typing.Coroutine')

    def is_list_type(self, typ: Type) -> bool:
        return (
            isinstance(typ, Instance) and typ.type.fullname == 'builtins.list'
        )

    def is_dict_call_type(self, expr: NameExpr) -> bool:
        # statically wise, TypedDicts do not inherit from dict
        # so we cannot check that, just checking if the expression
        # inherits from a class that ends with dict is good enough
        # for our use case
        return bool(expr.fullname == 'builtins.dict') or bool(
            isinstance(expr.node, TypeInfo)
            and expr.node.bases
            and expr.node.bases[0].type.fullname.lower().endswith('dict')
        )

    def copy_modified_instance(
        self, instance: Instance, names: SymbolTable
    ) -> Instance:
        new = copy.copy(instance)
        new.type = TypeInfo(names, new.type.defn, new.type.module_name)
        new.type.mro = [new.type, *instance.type.mro]
        new.type.bases = instance.type.bases
        new.type.metaclass_type = instance.type.metaclass_type
        return new

    def copy_modified_optional_type(
        self, original: UnionType, typ: Type
    ) -> UnionType:
        new = copy.copy(original)
        new.items = new.items.copy()
        new.items[0] = typ
        return new

    def parse_expression_to_dict(
        self, expression: Expression
    ) -> Dict[Any, Any]:
        if isinstance(expression, DictExpr):
            return self._dictexpr_to_dict(expression)

        if isinstance(expression, CallExpr):
            return self._callexpr_to_dict(expression)

        raise TypeError(
            f'Cannot parse expression of type={type(expression).__name__} to a dictionary.'
        )

    def _dictexpr_to_dict(self, expr: DictExpr) -> Dict[Any, Any]:
        parsed = {}
        for key_expr, value_expr in expr.items:
            if key_expr is None:
                # TODO: what causes this?
                continue

            key = self._resolve_expression(key_expr)
            value = self._resolve_expression(value_expr)
            parsed[key] = value

        return parsed

    def _callexpr_to_dict(
        self, expr: CallExpr, strict: bool = True
    ) -> Dict[str, Any]:
        if not isinstance(expr.callee, NameExpr):
            raise TypeError(
                f'Expected CallExpr.callee to be a NameExpr but got {type(expr.callee)} instead.'
            )

        if strict and not self.is_dict_call_type(expr.callee):
            raise TypeError(
                f'Expected builtins.dict to be called but got {expr.callee.fullname} instead'
            )

        parsed = {}
        for arg_name, value_expr in zip(expr.arg_names, expr.args):
            if arg_name is None:
                continue

            value = self._resolve_expression(value_expr)
            parsed[arg_name] = value

        return parsed

    def _resolve_expression(self, expression: Expression) -> Any:
        if isinstance(expression, (StrExpr, BytesExpr, IntExpr)):
            return expression.value

        if isinstance(expression, NameExpr):
            return self._resolve_name_expression(expression)

        if isinstance(expression, DictExpr):
            return self._dictexpr_to_dict(expression)

        if isinstance(expression, CallExpr):
            return self._callexpr_to_dict(expression)

        return expression

    def _resolve_name_expression(self, expression: NameExpr) -> Any:
        if isinstance(expression.node, Var):
            return self._resolve_var_node(expression.node)

        return expression

    def _resolve_var_node(self, node: Var) -> Any:
        if node.is_final:
            return node.final_value

        if node.fullname.startswith('builtins.'):
            return self._resolve_builtin(node.fullname)

        return node

    def _resolve_builtin(self, fullname: str) -> Any:
        return operator.attrgetter(*fullname.split('.')[1:])(builtins)


class UnparsedExpression(Exception):
    context: Union[Expression, Node]

    def __init__(self, context: Union[Expression, Node]) -> None:
        self.context = context
        super().__init__(
            f'Tried to access a ({type(context).__name__}) expression that was not parsed.'
        )


ERROR_PARSING = ErrorCode('prisma-parsing', 'Unable to parse', 'Prisma')


def error_unable_to_parse(
    api: CheckerPluginInterface, context: Context, detail: str
) -> None:
    link = (
        'https://github.com/RobertCraigie/prisma-client-py/issues/new/choose'
    )
    full_message = f'The prisma mypy plugin was unable to parse: {detail}\n'
    full_message += f'Please consider reporting this bug at {link} so we can try to fix it!'
    api.fail(full_message, context, code=ERROR_PARSING)



from __future__ import annotations

import asyncio
import logging
import os
from collections import defaultdict
from typing import Any, List, Optional, Type, TypeVar

try:
    import query as q

except ImportError:
    from . import query as q

from dotenv import load_dotenv
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from .client import FaunaClient
from .errors import FaunaException
from .json import JSONModel

load_dotenv()

T = TypeVar("T")

Model = Type[T]

MaybeModel = Optional[Model]

ModelList = List[Model]


class Fql(BaseModel):
    field: str

    operator: str

    value: Any


class FaunaModel(JSONModel):
    ref: Optional[str] = None

    ts: Optional[str] = None

    def __init__(self, **data: Any):
        super().__init__(**data)

        for field in self.__fields__.values():
            try:
                one_of = field.field_info.extra.get("oneOf")

                if isinstance(one_of, list):
                    if data.get(field.name) not in one_of:
                        raise ValueError(f"{field.name} must be one of {one_of}")

            except Exception:
                continue

    @classmethod
    def client(cls) -> FaunaClient:
        fauna_secret = os.getenv("FAUNA_SECRET")
        return FaunaClient(secret=fauna_secret)

    @classmethod
    def q(cls):
        return cls.client().query

    @classmethod
    async def provision(cls) -> bool:
        _q = cls.q()

        try:
            if not await _q(q.exists(q.collection(cls.__name__.lower()))):
                await _q(q.create_collection({"name": cls.__name__.lower()}))

                print("Created collection %s", cls.__name__.lower())

            if not await _q(q.exists(q.index(cls.__name__.lower()))):
                await _q(
                    q.create_index(
                        {
                            "name": cls.__name__.lower(),
                            "source": q.collection(cls.__name__.lower()),
                        }
                    )
                )

                print(f"Created index {cls.__name__.lower()}")

            for field in cls.__fields__.values():
                if field.field_info.extra.get("unique"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}_unique"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}_unique",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                    "unique": True,
                                }
                            )
                        )

                        print(cls.__name__.lower(), field.name)

                    print(
                        "Created unique index %s_%s", cls.__name__.lower(), field.name
                    )
                    continue

                if field.field_info.extra.get("index"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                }
                            )
                        )

                        print("Created index %s_%s", cls.__name__.lower(), field.name)
                        continue

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return False

    @classmethod
    async def exists(cls, ref: str) -> bool:
        try:
            return await cls.q()(
                q.exists(q.ref(q.collection(cls.__name__.lower()), ref))
            )

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return False

    @classmethod
    async def find_unique(cls, field: str, value: Any) -> MaybeModel:
        try:
            data = await cls.q()(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)
            return

    @classmethod
    async def find_many(cls, field: str, value: Any) -> ModelList:
        try:
            _q = cls.q()

            refs = (
                await _q(
                    q.paginate(
                        q.match(q.index(f"{cls.__name__.lower()}_{field}"), value)
                    )
                )
            )["data"]

            collection_refs_map = defaultdict(list)

            for ref in refs:
                collection = q.collection(cls.__name__.lower())

                collection_refs_map[collection].append(
                    q.ref(collection, ref["@ref"]["id"])
                )

            data = await asyncio.gather(
                *[
                    _q(q.get(q.paginate(collection_refs)))
                    for _, collection_refs in collection_refs_map.items()
                ]
            )

            flattened_data = [
                item for collection_data in data for item in collection_data["data"]
            ]

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in flattened_data
            ]

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return []

    @classmethod
    async def get(cls, ref: str) -> MaybeModel:
        try:
            data = await cls.q()(q.get(q.ref(q.collection(cls.__name__.lower()), ref)))
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)
            return

    @classmethod
    async def all(cls) -> ModelList:
        try:
            _q = cls.q()

            query = q.paginate(q.match(q.index(f"{cls.__name__.lower()}")))

            refs = (await _q(query))["data"]

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return []

    @classmethod
    async def delete_unique(cls, field: str, value: Any) -> bool:
        try:
            _q = cls.q()

            ref = await _q(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )

            await _q(q.delete(ref))

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return False

    @classmethod
    async def delete(cls, ref: str) -> bool:
        try:
            await cls.q()(q.delete(q.ref(q.collection(cls.__name__.lower()), ref)))

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return False

    async def create(self) -> MaybeModel:
        try:
            for field in self.__fields__.values():
                if field.field_info.extra.get("unique"):
                    instance = await self.find_unique(
                        field.name, self.__dict__[field.name]
                    )

                    if instance is None:
                        continue

                    if issubclass(instance.__class__, FaunaModel):
                        return instance

            data = await self.__class__.q()(
                q.create(
                    q.collection(self.__class__.__name__.lower()), {"data": self.dict()}
                )
            )

            self.ref = data["ref"]["@ref"]["id"]

            self.ts = data["ts"] / 1000
            return self

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)
            return

    @classmethod
    async def query(cls, query: str) -> ModelList:
        try:
            refs = (await cls.q()(q.paginate(q.match(q.query(query)))))["data"]

            data = await asyncio.gather(*[cls.q()(q.get(ref)) for ref in refs])

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            return []

    @classmethod
    async def update(cls, ref: str, **kwargs) -> MaybeModel:
        try:
            instance = await cls.get(ref)

            if isinstance(instance, cls):
                instance_updated = await cls.q()(
                    q.update(
                        q.ref(q.collection(cls.__name__.lower()), ref),
                        {"data": kwargs.get("kwargs", kwargs)},
                    )
                )
                return cls(
                    **{
                        **instance_updated["data"],
                        "ref": instance_updated["ref"]["@ref"]["id"],
                        "ts": instance_updated["ts"] / 1000,
                    }
                )

            else:
                raise ValueError(f"Field {ref} not found")

        except (FaunaException, KeyError, TypeError) as exc:
            logging.error(exc)

            raise ValueError(f"Field {ref} not found")

    async def save(self) -> MaybeModel:
        if isinstance(self.ref, str) and len(self.ref) == 18:
            return await self.update(self.ref, kwargs=self.dict())

        return await self.create()



from __future__ import annotations

import http.client
import http.server
import json
import urllib.parse
from typing import Any, Dict, Optional

from ..datastructures import LazyProxy


class HttpClient(LazyProxy[http.client.HTTPConnection]):
    """A simple HTTP client"""

    DEFAULT_HEADERS = {
        "Content-Type": "text/plain; charset=utf-8",
        "User-Agent": "Oscar Bahamonde <oscar.bahamonde.dev@gmail.com>",
    }

    def __load__(self) -> http.client.HTTPConnection:
        return http.client.HTTPConnection("localhost", 8080)

    def request(
        self, url: str, method: str, headers: Optional[Dict[str, str]], body: Any
    ):
        """Send a request to a URL"""
        parsed_url = urllib.parse.urlparse(url)
        conn = self._get_connection(parsed_url)
        headers = self._prepare_headers(headers)

        if headers.get("Content-Type") == "application/x-www-form-urlencoded":
            body = self._url_encode_body(body)
        if body is not None:
            conn.request(method, parsed_url.path, body.encode(), headers)
            with conn.getresponse() as res:
                response_headers = self._parse_response_headers(res)
                response_body = self._parse_response_body(res, response_headers)
                return {
                    "status": res.status,
                    "headers": response_headers,
                    "body": response_body,
                }
        else:
            conn.request(method, parsed_url.path, None, headers)
            with conn.getresponse() as res:
                response_headers = self._parse_response_headers(res)
                response_body = self._parse_response_body(res, response_headers)
                return {
                    "status": res.status,
                    "headers": response_headers,
                    "body": response_body,
                }

    def _url_encode_body(self, body: Dict[str, str]) -> str:
        """Encode the body as a URL-encoded string"""
        encoded_params = urllib.parse.urlencode(body)
        return encoded_params

    def get(self, url: str, headers: Optional[Dict[str, str]] = None):
        """Send a GET request to a URL"""
        return self.request(url, "GET", headers, None)

    def post(
        self, url: str, body: Dict[str, Any], headers: Optional[Dict[str, str]] = None
    ):
        """Send a POST request to a URL"""
        return self.request(url, "POST", headers, body)

    def put(self, url: str, body: Any, headers: Optional[Dict[str, str]] = None):
        """Send a PUT request to a URL"""
        return self.request(url, "PUT", headers, body)

    def delete(self, url: str, headers: Optional[Dict[str, str]] = None):
        """Send a DELETE request to a URL"""
        return self.request(url, "DELETE", headers, None)

    def _get_connection(self, parsed_url: urllib.parse.ParseResult):
        """Create an HTTPConnection or HTTPSConnection based on the URL scheme"""
        if parsed_url.scheme == "http":
            return http.client.HTTPConnection(parsed_url.netloc)
        elif parsed_url.scheme == "https":
            return http.client.HTTPSConnection(parsed_url.netloc)
        else:
            raise ValueError(f"Unsupported scheme: {parsed_url.scheme}")

    def _prepare_headers(self, headers: Optional[Dict[str, str]]):
        """Prepare headers with defaults"""
        prepared_headers = self.DEFAULT_HEADERS.copy()
        if headers:
            prepared_headers.update(headers)
        return prepared_headers

    def _parse_response_headers(self, response):
        """Parse response headers into a dictionary"""
        return {k.lower(): v for k, v in response.getheaders()}

    def _parse_response_body(self, response, headers):
        """Parse response body based on Content-Type header"""
        content_type = headers.get("content-type", "")
        if content_type.endswith("json"):
            return json.loads(response.read().decode())
        if content_type.startswith("text/"):
            return response.read().decode()
        if content_type.startswith("x-www-form-urlencoded"):
            return urllib.parse.parse_qs(response.read().decode())
        return response.read()
    
    
def main():
    """Run the HTTP client"""
    client = HttpClient()
    url = "http://localhost:8080"
    path = input("Enter path: ")
    method = input("Enter method: 1) GET 2) POST 3) PUT 4) DELETE: ")
    if method == "1":
        method = "GET"
    elif method == "2":
        method = "POST"
    elif method == "3":
        method = "PUT"
    elif method == "4":
        method = "DELETE"
    else:
        print("Invalid method")
        return
    response = client.request(url + path, method, None, None)
    print("Status: ", response["status"])
    print("Headers: ", response["headers"])
    print("Body: ", response["body"])
    print("Done")


from pydantic.fields import \
    Field as PydanticField  # pylint: disable=no-name-in-module


def Field(*args, index:bool=False, unique:bool=False, **kwargs):
    """Field Factory"""
    return PydanticField(*args, index=index, unique=unique, **kwargs)


from typing import Generic, Literal, TypeVar

from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

StatusCode = Literal[
    100,
    101,
    102,
    103,
    200,
    201,
    202,
    203,
    204,
    205,
    206,
    207,
    208,
    226,
    300,
    301,
    302,
    303,
    304,
    305,
    307,
    308,
    400,
    401,
    402,
    403,
    404,
    405,
    406,
    407,
    408,
    409,
    410,
    411,
    412,
    413,
    414,
    415,
    416,
    417,
    418,
    421,
    422,
    423,
    424,
    425,
    426,
    428,
    429,
    431,
    451,
    500,
    501,
    502,
    503,
    504,
    505,
    506,
    507,
    508,
    510,
    511,
]


Status = Literal["success", "error", "warning", "info"]

E = TypeVar("E", bound=Exception)


class AioFaunaException(Generic[E]):
    """Exception with status code, message, and status."""

    status_code: StatusCode
    message: str = Field(
        default="[The description from MDN]", description="The description from MDN"
    )
    status: Status = Field(default="error", description="The status of the exception")
    origin: str = Field(default="aiofauna", description="The origin of the exception")

    def json(self):
        return {"status": self.status, "message": self.message}

    def __repr__(self):
        return f"{self.__doc__}: {self.status_code} {self.message}"


class HttpException(AioFaunaException[Exception]):
    """Exception: An HTTP exception with status code, message, and status."""

    status_code: StatusCode
    message: str = Field(
        default="[The description from MDN]", description="The description from MDN"
    )
    status: Status = Field(default="error", description="The status of the exception")
    origin: str = Field(
        default="http server", description="The origin of the exception"
    )


class BadRequest(HttpException):
    """Exception: 400 Bad Request exception."""

    status_code: StatusCode = 400
    message: str = Field(default="Bad Request", description="Bad Request")
    status: Status = Field(default="error", description="The status of the exception")


class Unauthorized(HttpException):
    """Exception: 401 Unauthorized exception."""

    status_code: StatusCode = 401
    message: str = Field(default="Unauthorized", description="Unauthorized")
    status: Status = Field(default="error", description="The status of the exception")


class NotFound(HttpException):
    """Exception: 404 Not Found exception."""

    status_code: StatusCode = 404
    message: str = Field(default="Not Found", description="Not Found")
    status: Status = Field(default="error", description="The status of the exception")


class MethodNotAllowed(HttpException):
    """Exception: 405 Method Not Allowed exception."""

    status_code: StatusCode = 405
    message: str = Field(default="Method Not Allowed", description="Method Not Allowed")
    status: Status = Field(default="error", description="The status of the exception")


class InternalServerError(HttpException):
    """Exception: 500 Internal Server Error exception."""

    status_code: StatusCode = 500
    message: str = Field(
        default="Internal Server Error", description="Internal Server Error"
    )
    status: Status = Field(default="error", description="The status of the exception")


class NotImplemented_(HttpException):
    """Exception: 501 Not Implemented exception."""

    status_code: StatusCode = 501
    message: str = Field(default="Not Implemented", description="Not Implemented")
    status: Status = Field(default="error", description="The status of the exception")


class ServiceUnavailable(HttpException):
    """Exception: 503 Service Unavailable exception."""

    status_code: StatusCode = 503
    message: str = Field(
        default="Service Unavailable", description="Service Unavailable"
    )
    status: Status = Field(default="error", description="The status of the exception")


class GatewayTimeout(HttpException):
    """Exception: 504 Gateway Timeout exception."""

    status_code: StatusCode = 504
    message: str = Field(default="Gateway Timeout", description="Gateway Timeout")
    status: Status = Field(default="error", description="The status of the exception")


class TooManyRequests(HttpException):
    """Exception: 429 Too Many Requests exception."""

    status_code: StatusCode = 429
    message: str = Field(default="Too Many Requests", description="Too Many Requests")
    status: Status = Field(default="error", description="The status of the exception")


class Forbidden(HttpException):
    """Exception: 403 Forbidden exception."""

    status_code: StatusCode = 403
    message: str = Field(default="Forbidden", description="Forbidden")
    status: Status = Field(default="error", description="The status of the exception")


class Conflict(HttpException):
    """Exception: 409 Conflict exception."""

    status_code: StatusCode = 409
    message: str = Field(default="Conflict", description="Conflict")
    status: Status = Field(default="error", description="The status of the exception")


class UnprocessableEntity(HttpException):
    """Exception: 422 Unprocessable Entity exception."""

    status_code: StatusCode = 422
    message: str = Field(
        default="Unprocessable Entity", description="Unprocessable Entity"
    )
    status: Status = Field(default="error", description="The status of the exception")


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, Generic, Iterable, List, TypeVar, cast

from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

T = TypeVar("T")


class LazyProxy(Generic[T], ABC):
    """
    A LazyLoading proxy object that defers the loading of an object until it is accessed.
    It generates types dynamically, so it can be used as a base class for other classes.
    These classes will benefit from the lazy loading behavior which improves performance.
    Also, it can be used as a decorator for functions, which will be called when the function is called.
    Subclasses must implement the __load__ method to provide the logic for loading the proxied object.
    Usage:
    1. Subclass LazyProxy and implement the __load__ method.
    2. Accessing attributes, calling methods, or using other operations on the LazyProxy instance will trigger
         the loading of the proxied object.
    """

    def __init__(self) -> None:
        self.__proxied: T | None = None

    def __getattr__(self, attr: str) -> object:
        return getattr(self.__get_proxied__(), attr)

    def __repr__(self) -> str:
        return repr(self.__get_proxied__())

    def __dir__(self) -> Iterable[str]:
        return self.__get_proxied__().__dir__()

    def __get_proxied__(self) -> T:
        proxied = self.__proxied
        if proxied is not None:
            return proxied

        self.__proxied = proxied = self.__load__()
        return proxied

    def __set_proxied__(self, value: T) -> None:
        self.__proxied = value

    def __as_proxied__(self) -> T:
        """Helper method that returns the current proxy, typed as the loaded object"""
        return cast(T, self)

    @abstractmethod
    def __load__(self) -> T:
        ...


class Graph(BaseModel):

    """
    `Graph` is a class that represents a graph data structure. It contains a list of nodes and a list of edges. It also contains methods for adding nodes and edges to the graph, as well as methods for checking if the graph is cyclic and performing a topological sort on the graph.
    """

    nodes: List[int] = Field(default_factory=list)
    edges: List[List[int]] = Field(default_factory=list)

    def add_node(self, node: int):
        """
        The add_node method adds a node to the graph.
        """
        self.nodes.append(node)

    def add_edge(self, edge: List[int]):
        """
        The add_edge method adds an edge to the graph.
        """
        self.edges.append(edge)

    @property
    def is_cyclic(self) -> bool:
        """
        The is_cyclic property checks if the graph is cyclic.
        """

        return self._is_cyclic(self.nodes, self.edges)

    @staticmethod
    def _is_cyclic(nodes: List[int], edges: List[List[int]]) -> bool:
        """
        The _is_cyclic method iterates through each node in the graph and checks if the graph is cyclic using the _is_cyclic_helper method.
        """
        visited = set()
        for node in nodes:
            if node not in visited:
                if Graph._is_cyclic_helper(node, visited, edges):
                    return True
        return False

    @staticmethod
    def _is_cyclic_helper(node: int, visited: set, edges: List[List[int]]) -> bool:
        """
        The _is_cyclic_helper method performs a depth-first search on the node and its neighbors. If a node has already been visited, then the graph is cyclic.
        """
        if node in visited:
            return True
        visited.add(node)
        for edge in edges:
            if node in edge:
                if Graph._is_cyclic_helper(edge[1], visited, edges):
                    return True
        return False

    def topological_sort(self) -> List[int]:
        """
        The topological_sort method performs a topological sort on the graph using a modified depth-first search algorithm. It initializes a set of visited nodes and a list of sorted nodes, and then iterates through each node in the graph.
        """
        return self._topological_sort(self.nodes, self.edges)

    @staticmethod
    def _topological_sort(nodes: List[int], edges: List[List[int]]) -> List[int]:
        """
        _topological_sort_helper method to perform depth-first search on the node and its neighbors. Once all nodes have been visited, the method returns the sorted nodes.
        """
        visited = set()
        sorted_nodes = []
        for node in nodes:
            if node not in visited:
                Graph._topological_sort_helper(node, visited, edges, sorted_nodes)
        return sorted_nodes

    @staticmethod
    def _topological_sort_helper(
        node: int, visited: set, edges: List[List[int]], sorted_nodes: List[int]
    ):
        """
        The _topological_sort_helper method performs depth-first search on a node and its neighbors recursively. It adds the node to the visited set and then recursively visits its neighbors. Once all neighbors have been visited, it appends the node to the sorted list.
        """
        visited.add(node)
        for edge in edges:
            if node in edge:
                Graph._topological_sort_helper(edge[1], visited, edges, sorted_nodes)
        sorted_nodes.append(node)


class UploadFile(BaseModel):
    """
    File Upload Model
    """

    filename: str
    content_type: str
    data: bytes
    size: int
    headers: Dict[str, str]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiofauna',
 'aiofauna.cli',
 'aiofauna.prisma',
 'aiofauna.prisma.binaries',
 'aiofauna.prisma.cli',
 'aiofauna.prisma.cli.commands',
 'aiofauna.prisma.engine',
 'aiofauna.prisma.generator',
 'aiofauna.prisma_cleanup']

package_data = \
{'': ['*'],
 'aiofauna.cli': ['templates/*'],
 'aiofauna.prisma.generator': ['templates/*', 'templates/engine/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'Markdown>=3.4.3,<4.0.0',
 'Pygments>=2.15.1,<3.0.0',
 'aiohttp-devtools>=1.0.post0,<2.0',
 'aiohttp-sse>=2.1.0,<3.0.0',
 'aiohttp>=3.8.4,<4.0.0',
 'aiohttp_cors>=0.7.0,<0.8.0',
 'iso8601>=1.1.0,<2.0.0',
 'pydantic[dotenv]>=1.10.7,<2.0.0',
 'tomlkit>=0.11.8,<0.12.0']

entry_points = \
{'console_scripts': ['aiofauna = aiofauna.cli:main',
                     'aioprisma = aiofauna.prisma.cli:main']}

setup_kwargs = {
    'name': 'aiofauna',
    'version': '0.1.22',
    'description': '',
    'long_description': "---\ntitle: AioFauna\n---\n# AioFauna\n\n🚀 Introducing aiofauna: A full-stack framework built on top of Aiohttp, Pydantic and FaunaDB.\n\n🔥 Inspired by FastAPI focuses on Developer Experience, Productivity and Versatility.\n\n🌟 Features:\n\n✅ Supports Python 3.7+, comes with an opinionated ODM (Object Document Mapper) out of the box for FaunaDB that abstracts out complex FQL expressions into pythonic, fully typed asynchronous methods for all CRUD operations.\n\n✅ Powerful and Scalable: Being built on top of Aiohttp an asyncio based http server/client and FaunaDB an scalable serverless database for modern applications allows for powerful and seamless integrations.\n\n✅ Performant: As a framework built on top of Aiohttp it leverages the power of asyncio and the fastest python `HTTPClient` aiohttp plus the ubiquiness of FaunaDB to achieve high performance.\n\n✅ Automatic Swagger UI generation: Automatic generation of interactive Swagger UI documentation for instant testing of your `Api`.\n\n✅ SSE (Server Sent Events): Built-in support for SSE (Server Sent Events) for real-time streaming of data from FaunaDB to your application.\n\n✅ Websockets: Built-in support for Websockets for real-time bidirectional communication between your application and the resources served by AioFauna `Api`.\n\n✅ Robust data validation: Utilizes the rich features of Pydantic for data validation and serialization.\n\n✅ Auto-provisioning: Automatic management of indexes, unique indexes, and collections with `FaunaModel` ODM.\n\n✅ Full JSON communication: Focus on your data, don't worry about the communication protocol. Your `Api` will receive and return JSON.\n\n✅ Markdown and Html support with live reload: experiment an smooth frontend devserver experience without leaving your backend code with familiar syntax, `render_template` and `markdown_it` functions enable you to render static markdown files and html jinja templates with live reload.\n\n✅ Inspired by fastapi, you will work with almost the same syntax and features like path operations, path parameters, query parameters, request body, status codes, `/docs` automatic interactive API documentation, and decorated view functions and automatic serialization and deserialization of your data.\n\n💡 With aiofauna, you can build fast, scalable, and reliable modern applications, avoiding decision fatigue and focusing on what really matters, your data and your business logic.\n\n📚 Check out the aiofauna library, and start building your next-gen applications today! 🚀\n\n#Python #FaunaDB #Async #Pydantic #aiofauna\n\n⚙️ If you are using a synchronous framework check out [Faudantic](https://github.com/obahamonde/faudantic) for a similar experience with FaunaDB and Pydantic.\n\n📚 [Documentation](https://obahamonde-aiofauna-docs.smartpro.solutions) (Built with aiofauna)\n\n📦 [PyPi](https://pypi.org/project/aiofauna/)\n\n📦 [Demo](https://aiofauna-fwuw7gz7oq-uc.a.run.app/) (Whatsapp clone built with aiofauna)\n\n📦 [Swagger UI](https://aiofauna-fwuw7gz7oq-uc.a.run.app/docs) (Whatsapp clone built with aiofauna)\n\n📦 [GitHub](https://github.com/obahamonde/aiofauna)\n",
    'author': 'Oscar Bahamonde',
    'author_email': 'oscar.bahamonde.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/obahamonde/aiofauna',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Any

import aiohttp

from ._types import Method
from .http_abstract import AbstractResponse, AbstractHTTP, ASYNC_DEFAULT_CONFIG


__all__ = ('HTTP', 'Response', 'client')


class HTTP(AbstractHTTP[aiohttp.ClientSession, aiohttp.ClientResponse]):
    session: aiohttp.ClientSession
    session_kwargs = ASYNC_DEFAULT_CONFIG
    __slots__ = ()

    async def download(self, url: str, dest: str) -> None:
        async with self.session.get(url) as resp:
            resp.raise_for_status()
            with open(dest, 'wb') as fd:
                while True:
                    chunk = await resp.content.read(1024)
                    if not chunk:
                        break
                    fd.write(chunk)

    async def request(
        self, method: Method, url: str, **kwargs: Any
    ) -> 'Response':
        async with self.session.request(method, url, **kwargs) as resp:
            return Response(resp)
        
    def open(self) -> None:
        self.session = aiohttp.ClientSession(**self.session_kwargs)
        
    async def close(self) -> None:
        if self.should_close():
            await self.session.close()
            self.session = None # type: ignore[assignment]


client: HTTP = HTTP()


class Response(AbstractResponse[aiohttp.ClientResponse]):
    __slots__ = ()

    @property
    def status(self) -> int:
        return self.status

    async def json(self, **kwargs: Any) -> Any:
        return await self.original.json(**kwargs)

    async def text(self, **kwargs: Any) -> str:
        return await self.original.text(**kwargs)